package ru.dbotthepony.kstarbound.defs.tile

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableMap
import it.unimi.dsi.fastutil.objects.Object2DoubleMap
import it.unimi.dsi.fastutil.objects.Object2DoubleMaps
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.defs.AssetReference
import ru.dbotthepony.kstarbound.world.physics.CollisionType
import ru.dbotthepony.kstarbound.defs.ThingDescription

val Registry.Ref<TileDefinition>.isEmptyTile: Boolean
	get() = entry == BuiltinMetaMaterials.EMPTY

val Registry.Ref<TileDefinition>.isNullTile: Boolean
	get() = entry == BuiltinMetaMaterials.NULL || entry == null

val Registry.Ref<TileDefinition>.isObjectSolidTile: Boolean
	get() = entry == BuiltinMetaMaterials.OBJECT_SOLID

val Registry.Ref<TileDefinition>.isObjectPlatformTile: Boolean
	get() = entry == BuiltinMetaMaterials.OBJECT_PLATFORM

val Registry.Entry<TileDefinition>.isEmptyTile: Boolean
	get() = this == BuiltinMetaMaterials.EMPTY

val Registry.Entry<TileDefinition>.isNullTile: Boolean
	get() = this == BuiltinMetaMaterials.NULL

val Registry.Entry<TileDefinition>.isObjectSolidTile: Boolean
	get() = this == BuiltinMetaMaterials.OBJECT_SOLID

val Registry.Entry<TileDefinition>.isObjectPlatformTile: Boolean
	get() = this == BuiltinMetaMaterials.OBJECT_PLATFORM

val Registry.Entry<TileDefinition>.supportsModifiers: Boolean
	get() = !value.isMeta && value.supportsMods

fun Registry.Entry<TileDefinition>.supportsModifier(modifier: Registry.Entry<TileModifierDefinition>): Boolean {
	return !value.isMeta && value.supportsMods && !modifier.value.isMeta
}

fun Registry.Entry<TileDefinition>.supportsModifier(modifier: Registry.Ref<TileModifierDefinition>): Boolean {
	return !value.isMeta && value.supportsMods && modifier.isPresent && !modifier.value!!.isMeta
}

val Registry.Entry<LiquidDefinition>.isEmptyLiquid: Boolean
	get() = this == BuiltinMetaMaterials.NO_LIQUID

val Registry.Ref<LiquidDefinition>.isEmptyLiquid: Boolean
	get() = entry == null || entry == BuiltinMetaMaterials.NO_LIQUID

object BuiltinMetaMaterials {
	private fun make(id: Int, name: String, collisionType: CollisionType) = Registries.tiles.add(name, id, TileDefinition(
		materialId = id,
		materialName = "metamaterial:$name",
		descriptionData = ThingDescription.EMPTY,
		category = "meta",
		renderTemplate = AssetReference.empty(),
		renderParameters = RenderParameters.META,
		isMeta = true,
		supportsMods = false,
		collisionKind = collisionType,
		damageTable = AssetReference(TileDamageConfig(
			damageFactors = ImmutableMap.of(),
			damageRecovery = Double.MAX_VALUE,
			maximumEffectTime = 0.0,
			totalHealth = Double.MAX_VALUE,
			harvestLevel = Int.MAX_VALUE,
		))
	), isBuiltin = true)

	private fun makeMod(id: Int, name: String) = Registries.tileModifiers.add(name, id, TileModifierDefinition(
		modId = id,
		modName = "metamod:$name",
		descriptionData = ThingDescription.EMPTY,
		renderParameters = RenderParameters.META,
		isMeta = true,
		renderTemplate = AssetReference.empty(),
	), isBuiltin = true)

	/**
	 * air
	 */
	val EMPTY = make(65535, "empty", CollisionType.NONE)

	/**
	 * not set / out of bounds
	 */
	val NULL = make(65534, "null", CollisionType.NULL)

	val STRUCTURE = make(65533, "structure", CollisionType.BLOCK)
	val BIOME = make(65527, "biome", CollisionType.BLOCK)
	val BIOME1 = make(65528, "biome1", CollisionType.BLOCK)
	val BIOME2 = make(65529, "biome2", CollisionType.BLOCK)
	val BIOME3 = make(65530, "biome3", CollisionType.BLOCK)
	val BIOME4 = make(65531, "biome4", CollisionType.BLOCK)
	val BIOME5 = make(65532, "biome5", CollisionType.BLOCK)
	val BOUNDARY = make(65526, "boundary", CollisionType.SLIPPERY)
	val OBJECT_SOLID = make(65500, "objectsolid", CollisionType.BLOCK)
	val OBJECT_PLATFORM = make(65501, "objectplatform", CollisionType.PLATFORM)

	val MATERIALS: ImmutableList<Registry.Entry<TileDefinition>> = ImmutableList.of(
		EMPTY,
		NULL,
		STRUCTURE,
		BIOME,
		BIOME1,
		BIOME2,
		BIOME3,
		BIOME4,
		BIOME5,
		BOUNDARY,
		OBJECT_SOLID,
		OBJECT_PLATFORM,
	)

	val EMPTY_MOD = makeMod(65535, "empty")
	val BIOME_MOD = makeMod(65534, "biome")
	val UNDERGROUND_BIOME_MOD = makeMod(65533, "underground_biome")

	val NO_LIQUID = Registries.liquid.add("empty", 0, LiquidDefinition(
		name = "metaliquid:empty",
		liquidId = 0,
		color = RGBAColor.TRANSPARENT_BLACK,
		texture = "",
		isMeta = true,
		textureMovementFactor = 0.0,
		bottomLightMix = RGBAColor.TRANSPARENT_BLACK
	), isBuiltin = true)
}
