package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.objects.ObjectAVLTreeSet
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import it.unimi.dsi.fastutil.objects.ObjectArraySet
import kotlinx.coroutines.channels.Channel
import kotlinx.coroutines.future.await
import kotlinx.coroutines.launch
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import ru.dbotthepony.kstarbound.defs.tile.isEmptyTile
import ru.dbotthepony.kstarbound.defs.tile.isNullTile
import ru.dbotthepony.kstarbound.defs.tile.supportsModifier
import ru.dbotthepony.kstarbound.network.LegacyNetworkCellState
import ru.dbotthepony.kstarbound.network.packets.clientbound.TileDamageUpdatePacket
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.IChunkListener
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.api.TileColor
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity
import java.util.concurrent.CompletableFuture
import java.util.concurrent.locks.ReentrantLock
import java.util.function.Predicate
import kotlin.concurrent.withLock

class ServerChunk(world: ServerWorld, pos: ChunkPos) : Chunk<ServerWorld, ServerChunk>(world, pos) {
	/**
	 * Determines the state chunk is in, chunks are written to persistent storage
	 * only if they are in [FULL] state to avoid partially loaded or partially generated
	 * chunks from making its way into persistent storage.
	 */
	enum class State {
		FRESH, // Nothing is loaded

		TILES,
		MICRO_DUNGEONS,
		CAVE_LIQUID,
		TILE_ENTITIES,
		ENTITIES,

		FULL; // indicates everything has been loaded
	}

	var state: State = State.FRESH
		private set

	private var isBusy = false
	private var idleTicks = 0
	private var ticks = 0
	private val targetState = Channel<State>(Int.MAX_VALUE)
	private val permanent = ArrayList<Ticket>()
	private val temporary = ObjectAVLTreeSet<TimedTicket>()
	private var nextTicketID = 0
	// ticket lock because tickets *could* be canceled (or created) concurrently
	// BUT, front-end ticket creation in ServerWorld is expected to be called only on ServerWorld's thread
	// because ChunkMap is not thread-safe
	private val ticketsLock = ReentrantLock()
	private val loadJob = world.scope.launch { loadChunk() }

	var isUnloaded = false
		private set

	private suspend fun chunkGeneratorLoop() {
		while (true) {
			if (state == State.FULL)
				break

			val targetState = targetState.receive()

			while (state < targetState) {
				isBusy = true

				val nextState = ServerChunk.State.entries[state.ordinal + 1]

				try {
					when (nextState) {
						State.TILES -> {
							// tiles can be generated concurrently without any consequences
							CompletableFuture.runAsync(Runnable { prepareCells() }, Starbound.EXECUTOR).await()
						}

						State.MICRO_DUNGEONS -> {
							//LOGGER.error("NYI: Generating microdungeons for $chunk")
						}

						State.CAVE_LIQUID -> {
							//LOGGER.error("NYI: Generating cave liquids for $chunk")
						}

						State.TILE_ENTITIES -> {
							//LOGGER.error("NYI: Generating tile entities for $chunk")
						}

						State.ENTITIES -> {
							//LOGGER.error("NYI: Placing entities for $chunk")
						}

						else -> {}
					}

					bumpState(nextState)
				} catch (err: Throwable) {
					LOGGER.error("Exception while propagating $this to next generation state $nextState", err)
					break
				}
			}

			isBusy = false
		}

		isBusy = false
	}

	private suspend fun loadChunk() {
		try {
			val cells = world.storage.loadCells(pos).await()

			// very good.
			if (cells.isPresent) {
				loadCells(cells.value)
				bumpState(State.CAVE_LIQUID)

				world.storage.loadEntities(pos).await().ifPresent {
					for (obj in it) {
						obj.joinWorld(world)
					}
				}

				bumpState(State.FULL)
				isBusy = false
				return
			} else {
				// generate.
				chunkGeneratorLoop()
			}
		} catch (err: Throwable) {
			LOGGER.error("Exception while loading chunk $this", err)
		}
	}

	fun permanentTicket(target: State = State.FULL): ITicket {
		ticketsLock.withLock {
			return Ticket(target)
		}
	}

	fun temporaryTicket(time: Int, target: State = State.FULL): ITimedTicket {
		require(time > 0) { "Invalid ticket time: $time" }

		ticketsLock.withLock {
			return TimedTicket(time, target)
		}
	}

	interface ITicket {
		fun cancel()
		val isCanceled: Boolean
		val pos: ChunkPos
		val id: Int
		val chunk: CompletableFuture<ServerChunk>
		var listener: IChunkListener?
	}

	interface ITimedTicket : ITicket, Comparable<ITimedTicket> {
		val timeRemaining: Int
		fun prolong(ticks: Int)

		override fun compareTo(other: ITimedTicket): Int {
			val cmp = timeRemaining.compareTo(other.timeRemaining)
			if (cmp != 0) return cmp
			return id.compareTo(other.id)
		}
	}

	override fun cellChanges(x: Int, y: Int, cell: ImmutableCell) {
		super.cellChanges(x, y, cell)

		val permanent: List<Ticket>
		val temporary: List<TimedTicket>

		ticketsLock.withLock {
			permanent = ObjectArrayList(this.permanent)
			temporary = ObjectArrayList(this.temporary)
		}

		permanent.forEach { if (it.targetState <= state) it.listener?.onCellChanges(x, y, cell) }
		temporary.forEach { if (it.targetState <= state) it.listener?.onCellChanges(x, y, cell) }
	}

	private fun onTileHealthUpdate(x: Int, y: Int, isBackground: Boolean, health: TileHealth) {
		val permanent: List<Ticket>
		val temporary: List<TimedTicket>

		ticketsLock.withLock {
			permanent = ObjectArrayList(this.permanent)
			temporary = ObjectArrayList(this.temporary)
		}

		permanent.forEach { if (it.targetState <= state) it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
		temporary.forEach { if (it.targetState <= state) it.listener?.onTileHealthUpdate(x, y, isBackground, health) }
	}

	private abstract inner class AbstractTicket(val targetState: State) : ITicket {
		final override val id: Int = nextTicketID++
		final override val pos: ChunkPos
			get() = this@ServerChunk.pos

		final override var isCanceled: Boolean = false
		final override val chunk = CompletableFuture<ServerChunk>()

		init {
			isBusy = true

			if (this@ServerChunk.state >= targetState) {
				chunk.complete(this@ServerChunk)
			} else {
				this@ServerChunk.targetState.trySend(targetState)
			}
		}

		final override fun cancel() {
			if (isCanceled) return

			ticketsLock.withLock {
				if (isCanceled) return
				isCanceled = true
				chunk.cancel(false)
				listener = null
				cancel0()
			}
		}

		protected abstract fun cancel0()
		final override var listener: IChunkListener? = null
	}

	private inner class Ticket(state: State) : AbstractTicket(state) {
		init {
			permanent.add(this)
		}

		override fun cancel0() {
			permanent.remove(this)
		}
	}

	private inner class TimedTicket(expiresAt: Int, state: ServerChunk.State) : AbstractTicket(state), ITimedTicket {
		var expiresAt = expiresAt + ticks

		override val timeRemaining: Int
			get() = (expiresAt - ticks).coerceAtLeast(0)

		init {
			temporary.add(this)
		}

		override fun cancel0() {
			temporary.remove(this)
		}

		override fun prolong(ticks: Int) {
			if (ticks == 0 || isCanceled) return

			ticketsLock.withLock {
				if (isCanceled) return

				temporary.remove(this)
				expiresAt += ticks
				if (timeRemaining > 0) temporary.add(this)
			}
		}
	}

	private fun bumpState(newState: State) {
		if (newState == state) return

		require(newState >= state) { "Tried to downgrade $this state from $state to $newState" }

		if (newState >= State.ENTITIES) {
			this.state = State.FULL
		} else {
			this.state = newState
		}

		val permanent: List<Ticket>
		val temporary: List<TimedTicket>

		ticketsLock.withLock {
			permanent = ObjectArrayList(this.permanent)
			temporary = ObjectArrayList(this.temporary)
		}

		permanent.forEach { if (it.targetState <= state) it.chunk.complete(this) }
		temporary.forEach { if (it.targetState <= state) it.chunk.complete(this) }
	}

	fun copyCells(): Object2DArray<ImmutableCell> {
		if (cells.isInitialized()) {
			return Object2DArray(cells.value)
		} else {
			return Object2DArray(CHUNK_SIZE, CHUNK_SIZE, AbstractCell.EMPTY)
		}
	}

	data class DamageResult(val result: TileDamageResult, val health: TileHealth? = null, val stateBefore: AbstractCell? = null)

	fun damageTile(pos: Vector2i, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null): DamageResult {
		if (!cells.isInitialized()) {
			return DamageResult(TileDamageResult.NONE)
		}

		val cell = cells.value[pos.x, pos.y]

		if (cell.isIndestructible || cell.tile(isBackground).material.value.isMeta) {
			return DamageResult(TileDamageResult.NONE)
		}

		var damage = damage
		var result = TileDamageResult.NORMAL

		if (cell.dungeonId in world.protectedDungeonIDs) {
			damage = damage.copy(type = TileDamageType.PROTECTED)
			result = TileDamageResult.PROTECTED
		}

		val health = (if (isBackground) tileHealthBackground else tileHealthForeground).value[pos.x, pos.y]
		val tile = cell.tile(isBackground)

		val params = if (!damage.type.isPenetrating && tile.modifier.value.breaksWithTile) {
			tile.material.value.actualDamageTable + tile.modifier.value.actualDamageTable
		} else {
			tile.material.value.actualDamageTable
		}

		health.damage(params, sourcePosition, damage)
		onTileHealthUpdate(pos.x, pos.y, isBackground, health)

		if (health.isDead) {
			if (isBackground) {
				damagedTilesBackground.remove(pos)
			} else {
				damagedTilesForeground.remove(pos)
			}

			val copyHealth = health.copy()
			val mCell = cell.mutable()
			val mTile = mCell.tile(isBackground)

			mTile.material = BuiltinMetaMaterials.EMPTY
			mTile.color = TileColor.DEFAULT
			mTile.hueShift = 0f

			if (tile.modifier.value.breaksWithTile) {
				mTile.modifier = BuiltinMetaMaterials.EMPTY_MOD
			}

			setCell(pos.x, pos.y, mCell.immutable())
			health.reset()
			return DamageResult(result, copyHealth, cell)
		} else {
			if (isBackground) {
				damagedTilesBackground.add(pos)
			} else {
				damagedTilesForeground.add(pos)
			}

			return DamageResult(result, health, cell)
		}
	}

	private val damagedTilesForeground = ObjectArraySet<Vector2i>()
	private val damagedTilesBackground = ObjectArraySet<Vector2i>()

	fun tileDamagePackets(): List<TileDamageUpdatePacket> {
		val result = ArrayList<TileDamageUpdatePacket>()

		if (tileHealthBackground.isInitialized()) {
			val tileHealthBackground = tileHealthBackground.value

			for (x in 0 until CHUNK_SIZE) {
				for (y in 0 until CHUNK_SIZE) {
					val health = tileHealthBackground[x, y]

					if (!health.isHealthy) {
						result.add(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, true, health))
					}
				}
			}
		}

		if (tileHealthForeground.isInitialized()) {
			val tileHealthForeground = tileHealthForeground.value

			for (x in 0 until CHUNK_SIZE) {
				for (y in 0 until CHUNK_SIZE) {
					val health = tileHealthForeground[x, y]

					if (!health.isHealthy) {
						result.add(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, false, health))
					}
				}
			}
		}

		return result
	}

	override fun tick() {
		ticks++

		ticketsLock.withLock {
			while (temporary.isNotEmpty() && temporary.first().timeRemaining <= 0) {
				val ticket = temporary.first()
				ticket.isCanceled = true
				temporary.remove(ticket)
			}
		}

		var shouldUnload = !isBusy && temporary.isEmpty() && permanent.isEmpty()

		if (shouldUnload) {
			idleTicks++
			// don't load-save-load-save too frequently
			shouldUnload = idleTicks > 600
		} else {
			idleTicks = 0
		}

		if (shouldUnload) {
			unload()
			return
		}

		if (state != State.FULL)
			return

		super.tick()

		if (cells.isInitialized() && (damagedTilesBackground.isNotEmpty() || damagedTilesForeground.isNotEmpty())) {
			val tileHealthBackground = tileHealthBackground.value
			val tileHealthForeground = tileHealthForeground.value
			val cells = cells.value

			damagedTilesBackground.removeIf { (x, y) ->
				val health = tileHealthBackground[x, y]
				val result = !health.tick(cells[x, y].background.material.value.actualDamageTable)
				onTileHealthUpdate(x, y, true, health)
				result
			}

			damagedTilesForeground.removeIf { (x, y) ->
				val health = tileHealthForeground[x, y]
				val result = !health.tick(cells[x, y].foreground.material.value.actualDamageTable)
				onTileHealthUpdate(x, y, false, health)
				result
			}
		}
	}

	fun unload() {
		if (isUnloaded)
			return

		isUnloaded = true
		loadJob.cancel()
		targetState.close()

		if (state == State.FULL) {
			val unloadable = world.entityIndex
				.query(
					aabb,
					filter = Predicate { it.isApplicableForUnloading && aabb.isInside(it.position) },
					distinct = true, withEdges = false)

			world.storage.saveCells(pos, copyCells())
			world.storage.saveEntities(pos, unloadable)

			unloadable.forEach {
				it.remove()
			}
		}

		world.chunkMap.remove(pos)
	}

	fun cancelLoadJob() {
		loadJob.cancel()
	}

	fun legacyNetworkCells(): Object2DArray<LegacyNetworkCellState> {
		if (cells.isInitialized()) {
			val cells = cells.value
			return Object2DArray(width, height) { a, b -> cells[a, b].toLegacyNet() }
		} else {
			return Object2DArray(width, height, LegacyNetworkCellState.NULL)
		}
	}

	fun prepareCells() {
		val cells = cells.value

		for (x in 0 until width) {
			for (y in 0 until height) {
				val info = world.template.cellInfo(pos.tileX + x, pos.tileY + y)

				val state = cells[x, y].mutable()

				state.blockBiome = info.blockBiomeIndex
				state.envBiome = info.environmentBiomeIndex

				if (state.foreground.material.isNullTile) {
					state.foreground.material = info.foreground.entry ?: BuiltinMetaMaterials.EMPTY
					state.foreground.color = TileColor.DEFAULT
					state.foreground.hueShift = info.blockBiome?.hueShift(state.foreground.material) ?: 0f

					if (state.foreground.material.supportsModifier(info.foregroundMod)) {
						state.foreground.modifier = info.foregroundMod.entry ?: BuiltinMetaMaterials.EMPTY_MOD
						state.foreground.hueShift = info.blockBiome?.hueShift(info.foregroundMod) ?: 0f
					}
				}

				if (state.background.material.isNullTile) {
					state.background.material = info.background.entry ?: BuiltinMetaMaterials.EMPTY
					state.background.color = TileColor.DEFAULT
					state.background.hueShift = info.blockBiome?.hueShift(state.background.material) ?: 0f

					if (state.background.material.supportsModifier(info.backgroundMod)) {
						state.background.modifier = info.backgroundMod.entry ?: BuiltinMetaMaterials.EMPTY_MOD
						state.background.hueShift = info.blockBiome?.hueShift(info.backgroundMod) ?: 0f
					}
				}

				if (!state.foreground.material.isEmptyTile) {
					state.liquid.reset()
				} else if (y < info.oceanLiquidLevel && info.oceanLiquid.isPresent && !info.oceanLiquid.value!!.isMeta) {
					val pressure = (info.oceanLiquidLevel - y).toFloat()

					if (state.background.material.isEmptyTile) {
						state.liquid.pressure = pressure
						state.liquid.isInfinite = true
						state.liquid.state = info.oceanLiquid.entry!!
					} else if (info.encloseLiquids) {
						state.liquid.pressure = pressure
						state.liquid.isInfinite = false
						state.liquid.level = 1f
						state.liquid.state = info.oceanLiquid.entry!!
					}
				}

				cells[x, y] = state.immutable()
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()
	}
}
